/*
 * 01/25/2003 - 14:02:55
 *
 * Console.java - A console (emulates system terminal)
 * Copyright (C) 1999-2003 Romain Guy
 * romain.guy@jext.org
 * www.jext.org
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.jext.console;

import java.io.*;
import java.net.*;

import java.awt.*;
import java.awt.event.*;

import javax.swing.*;
import javax.swing.text.Document;
import javax.swing.text.StyleConstants;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.BadLocationException;

import org.jext.*;
import org.jext.console.commands.*;

/**
 * An internal console which provide different kinds of
 * prompts and which allows to execute both internal and
 * external (OS specific) commands. The console is embedded
 * in a <code>JScrollPane</code> and handles it by itself.
 * @author Romain Guy
 */

public class Console extends JScrollPane
{
  /** DOS prompt: /export/home/guy > */
  public static final int DOS_PROMPT = 0;
  /** Jext prompt: Gfx@/export/home/guy > */
  public static final int JEXT_PROMPT = 1;
  /** Linux prompt: guy@csdlyon$ */
  public static final int LINUX_PROMPT = 2;
  /** SunOS prompt: csdlyon% */
  public static final int SUNOS_PROMPT = 3;

  /** Default prompt types: DOS, Jext, Linux and SunOS **/
  public static final String[] DEFAULT_PROMPTS = { "$p >", "$u@$p >", "$u@$h$$ ", "$h% " };

  // current separators used in command lines
  private static final String COMPLETION_SEPARATORS = " \t;:/\\\"\'";

  // commands
  private Command currentCmd, firstCmd;
  // parent
  private JextFrame parent;
  // processes specific
  private Process process;
  private String processName;
  private StdoutThread stdout;
  private StderrThread stderr;
  // private fields
  private String current;
  private Document outputDocument;
  private ConsoleTextPane textArea;
  private HistoryModel historyModel = new HistoryModel(25);
  private int userLimit = 0, typingLocation = 0, index = -1;
  // colors
  public Color errorColor = Color.red;
  public Color promptColor = Color.blue;
  public Color outputColor = Color.black;
  public Color infoColor = new Color(0, 165, 0);
  // prompt
  private boolean displayPath;
  private String prompt, hostName, oldPath = new String();
  private String promptPattern = DEFAULT_PROMPTS[JEXT_PROMPT];

  private Command evalCom = null;
  
  /**
   * Instanciates a new console without displaying prompt.
   * @param parent <code>Jext</code> parent
   */

  public Console(JextFrame parent)
  {
    this(parent, false);
  }

  /**
   * Creates a new console, embedding it in a <code>JScrollPane</code>.
   * By default console help is displayed.
   * @param parent <code>Jext</code> parent
   * @param display If set on true, prompt is displayed
   */

  public Console(JextFrame parent, boolean display)
  {
    super(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS,
          ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);

    this.parent = parent;

    // load commands from previous sessions
    load();

    textArea = new ConsoleTextPane(this);
    textArea.setFont(new Font("Monospaced", Font.PLAIN, 11));
    outputDocument = textArea.getDocument();
    append(Jext.getProperty("console.welcome"), infoColor, false, true);

    if (display)
      displayPrompt();

    getViewport().setView(textArea);
    FontMetrics fm = getFontMetrics(textArea.getFont());
    setPreferredSize(new Dimension(40 * fm.charWidth('m'), 6 * fm.getHeight()));
    setMinimumSize(getPreferredSize());
    setMaximumSize(getPreferredSize());

    initCommands();
  }

  /**
   * Returns this Console's parent.
   */

  public JextFrame getParentFrame()
  {
    return parent;
  }

  /**
   * Return the <code>Document</code> in which output is performed.
   */

  public Document getOutputDocument()
  {
    return outputDocument;
  }

  /**
   * Adds a command to the linked list of commands.
   */

  public void addCommand(Command command)
  {
    if (command == null)
      return;

    currentCmd.next = command;
    currentCmd = command;
  }

  /**
   * Return true if command is built-in. If command is built-in,
   * it is also executed.
   * @param command Command to check and execute
   */

  private boolean builtInCommand(String command)
  {
    boolean ret = false;
    Command _currentCmd = firstCmd;

    while (_currentCmd != null)
    {
      if (_currentCmd.handleCommand(this, command))
      {
        ret = true;
        break;
      }
      _currentCmd = _currentCmd.next;
    }

    return ret;
  }

  // inits commands list

  private void initCommands()
  {
    firstCmd = currentCmd = new ClearCommand();
    addCommand(new JythonCommand());
    addCommand(new ChangeDirCommand());
    addCommand(new ExitCommand());
    addCommand(new FileCommand());
    addCommand(new HomeCommand());
    addCommand(new HttpCommand());
    addCommand(new HelpCommand());
    addCommand(new ListCommand());
    addCommand(new PwdCommand());
    addCommand(new RunCommand());
    addCommand(evalCom = new EvalCommand());
  }

  /**
   * Set console background color.
   * @param color <code>Color</code> to be used
   */

  public void setBgColor(Color color)
  {
    textArea.setBackground(color);
  }

  /**
   * Set console error color.
   * @param color <code>Color</code> to be used
   */

  public void setErrorColor(Color color)
  {
    errorColor = color;
  }

  /**
   * Set console prompt color.
   * @param color <code>Color</code> to be used
   */

  public void setPromptColor(Color color)
  {
    promptColor = color;
  }

  /**
   * Set console output color.
   * @param color <code>Color</code> to be used
   */

  public void setOutputColor(Color color)
  {
    outputColor = color;
    textArea.setForeground(color);
    textArea.setCaretColor(color);
  }

  /**
   * Set console info color.
   * @param color <code>Color</code> to be used
   */

  public void setInfoColor(Color color)
  {
    infoColor = color;
  }

  /**
   * Set console selection color.
   * @param color <code>Color</code> to be used
   */

  public void setSelectionColor(Color color)
  {
    textArea.setSelectionColor(color);
  }

  /**
   * Save the history.
   */

  public void save()
  {
    for (int i = 0; i < historyModel.getSize(); i++)
      Jext.setProperty("console.history." + i, historyModel.getItem(i));
  }

  /**
   * Load the last saved history.
   */

  public void load()
  {
    String s;

    for (int i = 24 ; i >= 0; i--)
    {
      s = Jext.getProperty("console.history." + i);
      if (s != null)
        historyModel.addItem(s);
    }
  }

  /**
   * Set the prompt pattern.
   * @param type The prompt pattern
   */

  public void setPromptPattern(String prompt)
  {
    if (prompt == null)
      return;

    promptPattern = prompt;
    displayPath = false;
    buildPrompt();
  }

  /**
   * Get prompt pattern.
   */

  public String getPromptPattern()
  {
    return promptPattern;
  }

  /**
   * Displays the prompt according to the current selected
   * prompt type.
   */

  public void displayPrompt()
  {
    if (prompt == null || displayPath)
      buildPrompt();

    if (Jext.getBooleanProperty("console.jythonMode"))
      append('\n' + "[python] " + prompt, promptColor);
    else
      append('\n' + prompt, promptColor);
    typingLocation = userLimit = outputDocument.getLength();
  }

  // builds the prompt according to the prompt pattern

  private void buildPrompt()
  {
    if (displayPath && oldPath.equals(System.getProperty("user.dir")))
      return;

    displayPath = false;
    StringBuffer buf = new StringBuffer();

    if (hostName == null)
    {
      try
      {
        hostName = InetAddress.getLocalHost().getHostName();
      } catch (UnknownHostException uhe) { }
    }

    for (int i = 0; i < promptPattern.length(); i++)
    {
      char c = promptPattern.charAt(i);

      switch(c)
      {
        case '$':
          if (i == promptPattern.length() - 1)
            buf.append(c);
          else
          {
            switch (promptPattern.charAt(++i))
            {
              case 'p':                    // current path
                buf.append(oldPath = System.getProperty("user.dir"));
                displayPath = true;
                break;
              case 'u':                    // user name
                buf.append(System.getProperty("user.name"));
                break;
              case 'h':                    // host name
                buf.append(hostName);
                break;
              case '$':
                buf.append('$');
                break;
            }
          }
          break;
        default:
          buf.append(c);
      }
    }

    prompt = buf.toString();
  }

  /**
   * This method appends text in the text area.
   * @param text The text to append
   * @param color The color of the text
   * @param italic Set to true will append italic text
   * @param bold Set to true will append bold text
   */

  private void append(String text, Color color, boolean italic, boolean bold)
  {
    SimpleAttributeSet style = new SimpleAttributeSet();
    if (color != null)
      style.addAttribute(StyleConstants.Foreground, color);
    StyleConstants.setBold(style, bold);
    StyleConstants.setItalic(style, italic);

    try
    {
      outputDocument.insertString(outputDocument.getLength(), text, style);
    } catch(BadLocationException bl) { }

    textArea.setCaretPosition(outputDocument.getLength());
  }

  /**
   * This method appends text in the text area.
   * @param text The text to append in the text area
   * @apram color The color of the text to append
   */

  public void append(String text, Color color)
  {
    append(text, color, false, false);
  }

  /**
   * Adds a command to the history.
   * @param command Command to add in the history
   */

  public void addHistory(String command)
  {
    historyModel.addItem(command);
    index = -1;
  }

  /**
   * Remove a char from current command line.
   * Stands for BACKSPACE action.
   */

  public void removeChar()
  {
    try
    {
      //if (typingLocation != userLimit)
      //  outputDocument.remove(--typingLocation, 1);
      int pos = textArea.getCaretPosition();
      if (pos <= typingLocation && pos > userLimit)
      {
        outputDocument.remove(pos - 1, 1);
        typingLocation--;
      }
    } catch (BadLocationException ble) { }
  }

  /**
   * Delete a char from command line.
   * Stands for DELETE action.
   */

  public void deleteChar()
  {
    try
    {
      int pos = textArea.getCaretPosition();
      if (pos == outputDocument.getLength()) return;
      if (pos < typingLocation && pos >= userLimit)
      {
        outputDocument.remove(pos, 1);
        typingLocation--;
      }
    } catch (BadLocationException ble) { }
  }

  /**
   * Adds a <code>String</code> to the current command line.
   * @param add <code>String</code> to be added
   */

  public void add(String add)
  {
    try
    {
      int pos = textArea.getCaretPosition();
      if (pos <= typingLocation && pos >= userLimit)
        outputDocument.insertString(pos, add, null);
      typingLocation += add.length();
    } catch (BadLocationException ble) { }
  }

  /**
   * Returns the position in characters at which
   * user is allowed to type his commands.
   * @return Beginning of user typing space
   */

  public int getUserLimit()
  {
    return userLimit;
  }

  /**
   * Returns the position of the end of the console prompt.
   */

  public int getTypingLocation()
  {
    return typingLocation;
  }

  /**
   * Completes current filename if possible.
   */

  public void doCompletion()
  {
    int index = 0;
    int caret = textArea.getCaretPosition() - userLimit;

    String wholeText = getText();
    String text;

    try
    {
      text = outputDocument.getText(userLimit, caret);
    } catch (BadLocationException ble) { return; }

    for (int i = text.length() - 1; i >= 0; i--)
    {
      if (COMPLETION_SEPARATORS.indexOf(text.charAt(i)) != -1)
      {
        if (i == index)
          return;
        index = i + 1;
        break;
      }
    }

    String current = text.substring(index);

    java.util.ArrayList matching = new java.util.ArrayList();
    String[] files = Utilities.getWildCardMatches("*", true);
    if (files == null)
      return;

    for (int i = 0; i < files.length; i++)
    {
      if (files[i].startsWith(current))
      {
        matching.add(files[i]);
      }
    }

    if (matching.size() == 0)
      return;

    int length = 0;
    int _length = 0;
    int mIndex = 0;

    for (int i = 0; i < matching.size(); i++)
    {
      _length = ((String) matching.get(i)).length();
      length = length < _length ? _length : length;
      if (length == _length)
        mIndex = i;
    }

    char c;
    boolean isSame = true;
    int diffIndex = length;

    String compare;
    String source = (String) matching.get(mIndex);

    for (int i = 0; i < length; i++)
    {
      c = source.charAt(i);

      for (int j = 0; j < matching.size(); j++)
      {
        if (j == mIndex)
          continue;

        compare = (String) matching.get(j);
        if (i >= compare.length())
          continue;

        isSame = (compare.charAt(i) == c);
      }

      if (!isSame)
      {
        diffIndex = i;
        break;
      }
    }

    compare = text.substring(0, index) + source.substring(0, diffIndex);
    setText(compare +  wholeText.substring(caret));
    textArea.setCaretPosition(userLimit + compare.length());
  }

  /**
   * Search backward in the history for a matching command,
   * according to the command typed in the user typing space.
   */

  public void doBackwardSearch()
  {
    String text = getText();
    if (text == null)
    {
      historyPrevious();
      return;
    }

    for(int i = index + 1; i < historyModel.getSize(); i++)
    {
      String item = historyModel.getItem(i);
      if (item.startsWith(text))
      {
        setText(item);
        index = i;
        return;
      }
    }
  }

  /**
   * Get previous item in the history list.
   */

  public void historyPrevious()
  {
    if (index == historyModel.getSize() - 1)
      getToolkit().beep();
    else if (index == -1)
    {
      current = getText();
      setText(historyModel.getItem(0));
      index = 0;
    } else {
      int newIndex = index + 1;
      setText(historyModel.getItem(newIndex));
      index = newIndex;
    }
  }

  /**
   * Get next item in the history list.
   */

  public void historyNext()
  {
    if (index == -1)
      getToolkit().beep();
    else if (index == 0)
      setText(current);
    else
    {
      int newIndex = index - 1;
      setText(historyModel.getItem(newIndex));
      index = newIndex;
    }
  }

  /**
   * Set user's command line content.
   * @param text Text to be put on command line.
   */

  public void setText(String text)
  {
    try
    {
      outputDocument.remove(userLimit, typingLocation - userLimit);
      outputDocument.insertString(userLimit, text, null);
      typingLocation = outputDocument.getLength();
      index = -1;
    } catch (BadLocationException ble) { }
  }

  /**
   * Returns current command line.
   */

  public String getText()
  {
    try
    {
      return outputDocument.getText(userLimit, typingLocation - userLimit);
    } catch (BadLocationException ble) { }
    return null;
  }

  /**
   * Display a message using output color.
   * @param display <code>String</code> to be displayed
   */

  public void output(String display)
  {
    append('\n' + display, outputColor, false, false);
  }

  /**
   * Displays console help.
   */

  public void help()
  {
    Command _current = firstCmd;
    StringBuffer buf = new StringBuffer();

    while (_current != null)
    {
      buf.append("   - ").append(_current.getCommandName());
      buf.append(Utilities.createWhiteSpace(30 - _current.getCommandName().length())).append('(');
      buf.append(_current.getCommandSummary()).append(')').append('\n');
      _current = _current.next;
    }

    help(Jext.getProperty("console.help", new String[] { buf.toString() }));
  }

  /**
   * Display a message using help color.
   * @param display <code>String</code> to be displayed
   */

  public void help(String display)
  {
    append('\n' + display, infoColor, true, true);
  }

  /**
   * Display a message using error color.
   * @param display <code>String</code> to be displayed
   */

  public void error(String display)
  {
    append('\n' + display, errorColor, false, false);
  }

  /**
   * Stops current task.
   */

  public void stop()
  {
    if (stdout != null)
    {
      stdout.interrupt();
      stdout = null;
      //stdout.stop();
    }

    if (stderr != null)
    {
      stderr.interrupt();
      stderr = null;
      //stderr.stop();
    }

    if (process != null)
    {
      process.destroy();
      Object[] args = { processName };
      error(Jext.getProperty("console.killed", args));
      processName = null;
      process = null;
    }
  }

  /**
   * Parse a command. Replace internal variables by their
   * values.
   * @param command Command to be parsed
   */

  public String parseCommand(String command)
  {
    String file;
    StringBuffer buf = new StringBuffer();
    for (int i = 0; i < command.length(); i++)
    {
      char c = command.charAt(i);
      switch(c)
      {
        case '$':
          if (i == command.length() - 1)
            buf.append(c);
          else
          {
            switch (command.charAt(++i))
            {
              case 'f':                    // current opened file (absolute path)
                file = parent.getTextArea().getCurrentFile();
                if (file != null)
                  buf.append(file);
                break;
              case 'd':                    // user directory
                buf.append(Utilities.getUserDirectory());
                break;
              case 'p':                    // current opened file name
                buf.append(parent.getTextArea().getName());
                break;
              case 'e':                    // current opened file name without extension
                file = parent.getTextArea().getName();
                int index = file.lastIndexOf('.');
                if (index != -1 && index + 1 < file.length())
                  buf.append(file.substring(0, index));
                else
                  buf.append(file);
                break;
              case 'n':                    // current opened file directory
                file = parent.getTextArea().getCurrentFile();
                if (file != null)
                  buf.append(file.substring(0, file.lastIndexOf(File.separator)));
                break;
              case 'h':                    // home dir
                buf.append(Utilities.getHomeDirectory());
                break;
              case 'j':                    // jext dir
                buf.append(Jext.getHomeDirectory());
                break;
              case 's':                    // selected text
                buf.append(parent.getTextArea().getSelectedText());
                break;
              case '$':
                buf.append('$');
                break;
            }
          }
          break;
        default:
          buf.append(c);
      }
    }
    return buf.toString();
  }

  /**
   * Execute command. First parse it then check if command
   * is built-in. At last, a process is created and threads
   * which handle output streams are started.
   * @param command Command to be execute
   */

  public void execute(String command)
  {
    stop();
    if (command.length() == 0 || command == null)
      return;

    // check to see if in "jython" mode...
    boolean isJython = Jext.getBooleanProperty("console.jythonMode");

    // if in jython mode, look for '!' as first charater...means
    // treat it like a "normal console" command
    if (isJython)
    {
      if (!command.startsWith("!"))
      {
        if (command.startsWith("?"))
        {  // kludge shorthand for 'print'
          String ts = command.substring(1);
          command = "print " + ts;
        } else if (command.startsWith("exit")) {
          Jext.setProperty("console.jythonMode", "off");
          displayPrompt();
          return;
        }

        evalCom.handleCommand(this, "eval:" + command);
        displayPrompt();
        return;

      } else {
        String ts = command.substring(1);
        command = ts;
      }
    }

    int index = command.indexOf(' ');
    if (index != -1)
      processName = command.substring(0, index);
    else
      processName = command;

    command = parseCommand(command);
    if (command == null || command.length() == 0)
      return;

    if (builtInCommand(command))
    {
      displayPrompt();
      return;
    }

    append("\n> " + command, infoColor);
    try
    {
      if (Utilities.JDK_VERSION.charAt(2) < '3')
        process = Runtime.getRuntime().exec(command);
      else
        process = Runtime.getRuntime().exec(command, null, new File(System.getProperty("user.dir")));
      process.getOutputStream().close();
    } catch (IOException ioe) {
      error(Jext.getProperty("console.error"));
      displayPrompt();
      return;
    }

    stdout = new StdoutThread();
    stderr = new StderrThread();

    if (process == null)
      displayPrompt();
  }

  class StdoutThread extends Thread
  {
    StdoutThread()
    {
      super("----thread: stout: jext----");
      start();
    }

    public void run()
    {
      try
      {
        BufferedReader in = new BufferedReader(new InputStreamReader(process.getInputStream()));

        String line;
        while((line = in.readLine()) != null)
          output(line);
        in.close();

        int exitCode = process.waitFor();
        Object[] args = { processName, new Integer(exitCode) };
        append('\n' + Jext.getProperty("console.exited", args), infoColor);

        Thread.sleep(500);
        process.destroy();
        processName = null;
        process = null;

        displayPrompt();
      } catch(IOException io) {
      } catch(InterruptedException ie) {
      } catch (NullPointerException npe) { }
    }
  }

  class StderrThread extends Thread
  {
    StderrThread()
    {
      super("----thread: stderr: jext----");
      start();
    }

    public void run()
    {
      try
      {
        if (process == null)
          return;
        BufferedReader in = new BufferedReader(new InputStreamReader(process.getErrorStream()));

        String line;
        while((line = in.readLine()) != null)
          append('\n' + line, errorColor);
        in.close();
      } catch(IOException io) {
      } catch (NullPointerException npe) { }
    }
  }

  //////////////////////////////////////////////////////////////////////////////////////////////
  // NEEDED BY JavaScriptParser PLUGIN
  //////////////////////////////////////////////////////////////////////////////////////////////

  private Writer writerSTDOUT = new Writer()
  {
    public void close() { }

    public void flush()
    {
      repaint();
    }

    public void write( char cbuf[], int off, int len )
    {
      append(new String(cbuf, off, len), outputColor);
    }
  };

  private Writer writeSTDERR = new Writer()
  {
    public void close() { }

    public void flush()
    {
      repaint();
    }

    public void write( char cbuf[], int off, int len )
    {
      append(new String(cbuf, off, len), errorColor);
    }
  };

  /**
   * Returns a writer in which external classes can send
   * <code>String</code> to make them being displayed in the
   * console as standard output.
   */

  public Writer getStdOut()
  {
    return writerSTDOUT;
  }

  /**
   * Returns a writer in which external classes can send
   * <code>String</code> to make them being displayed in the
   * console as error output.
   */

  public Writer getStdErr()
  {
    return writeSTDERR;
  }
  
  /***************************************************************************
  Patch
     -> Memory management improvements : it may help the garbage collector.
     -> Author : Julien Ponge (julien@izforge.com)
     -> Date : 23, May 2001
  ***************************************************************************/
  protected void finalize() throws Throwable
  {
    super.finalize();
    
    currentCmd = null;
    firstCmd = null;
    
    parent = null;
    
    process = null;
    processName = null;
    stdout = null;
    stderr = null;
    
    current = null;
    outputDocument = null;
    textArea = null;
    historyModel = null;
    
    infoColor = null;
    
    prompt = null;
    hostName = null;
    oldPath = null;          
    promptPattern = null;
  }
  // End of patch
}

// End of Console.java
