/*
 * 03/23/2003
 *
 * PluginDownload.java - The manager for plugins update.
 * Copyright (C) 2003 Paolo Giarrusso
 * blaisorblade_work@yahoo.it
 * www.jext.org
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package org.jext.misc;

import java.net.*;
import java.io.*;
import org.jext.Jext;
import org.jext.JARClassLoader;

public class PluginDownload {
  private PluginDownload() {}//"static" class

  private static URL autoUpdateVersionUrl;

  private final static String jarName = "autoUpdate.jar";
  private static URL autoUpdateJarUrl;
  private final static File downloadedJarPath = new File(Jext.SETTINGS_DIRECTORY + jarName);
  private static String defaultJarPath;
  private static ClassLoader loader = null, newVerLoader = null, defLoader = null;

  static {
    try {
      autoUpdateVersionUrl = new URL("http://www.jext.org/plugReader.version");
      autoUpdateJarUrl = new URL("http://www.jext.org/" + jarName);
    } catch (MalformedURLException mue) {}
  }

  private static String getDefaultJarPath() {
    if (defaultJarPath == null)
      defaultJarPath = Jext.JEXT_HOME + File.separator + ".." + File.separator + 
        "lib" + File.separator + jarName;
    return defaultJarPath;
  }

  public static void downloadBoot() {
    try {
      InputStream releaseInp = autoUpdateVersionUrl.openStream();
      byte buf[] = new byte[10];
      releaseInp.read(buf);
      int currVersion = Integer.parseInt(Jext.getProperty("plugDownload.core.version")),
          newVersion = Integer.parseInt(new String(buf));
      if (currVersion < newVersion) {
        InputStream jarInp = new BufferedInputStream(autoUpdateJarUrl.openStream());
        OutputStream saveStream = new BufferedOutputStream(new FileOutputStream(downloadedJarPath));
        buf = new byte[500];
        int nRead;
        while ((nRead = jarInp.read(buf)) != -1) {
          saveStream.write(buf, 0, nRead);
        }
        jarInp.close();
        saveStream.close();
      }
    } catch (IOException ioe) {
      //In this case, we can't update the autoUpdate.jar file; but then we use the current one.
      System.err.println("Caught exception while trying to update autoUpdate.jar");
      ioe.printStackTrace();
    }
    //Now, if autoUpdate.jar has been updated, it is in the right place in the user's home;
    //otherwise we need to use the default one, in jext/lib dir. Or maybe somewhere else, such as bin,
    //since it mustn't be loaded at startup.
    //And we build the chaining class loader that will be used to load all resources.
    buildChainingClassLoader();
  }

  private static boolean buildChainingClassLoader() {

    //We build first a ClassLoader which loads the supplied autoUpdate.jar, then
    //another one which looks in the downloaded one BUT uses as fallback one
    //the first one. The fallback is provided trasparently by the JDK by using
    //the fallback classloader as parent of the new one. But this only works for
    //resources, since for classes we must fallback not only in the case of ClassNotFoundEx.
    try {
      defLoader = new JARClassLoader(getDefaultJarPath(), false, null);
    } catch (IOException ioe) {
      System.err.println("You haven't installed correctly Jext! The autoUpdate.jar file is missing." +
          "It should be in this position: " + getDefaultJarPath());
    }

    if (downloadedJarPath.exists())
      try {
        newVerLoader = new JARClassLoader(downloadedJarPath.getName(), false, defLoader);
        //here defLoader becomes the father of newVerLoader
      } catch (IOException ioe) {
        //The file is there, but there could be problems anyway.
      }
    if (newVerLoader != null)
      loader = newVerLoader;
    else if (defLoader != null)
      loader = defLoader;
    else
      return false;
    return true;
  }

  private static Object getInstanceFromLoader(String className) {
    if (loader != null)
      try {
        return loader.loadClass(className).newInstance();
      } catch (InstantiationException ie) {
      } catch (IllegalAccessException ie) {
      } catch (ClassNotFoundException ie) {
        return null;//the class is missing in both chained classLoaders.
      }

    //if the class was found in newVerLoader, we've tried with it.
    //So we only need trying with defLoader, and only if != loader.
    if (defLoader != null && defLoader != loader)
      try {
        return defLoader.loadClass(className).newInstance();
      } catch (InstantiationException ie) {
      } catch (IllegalAccessException ie) {
      } catch (ClassNotFoundException ie) {
      }
    return null;
  }

  public static AbstractPlugReader getUpdater() {
    return (AbstractPlugReader) getInstanceFromLoader("PlugReader");
  }

  public static AbstractChoiceForm getUpdateWindow() {
    return (AbstractChoiceForm) getInstanceFromLoader("ChoiceForm");
  }

  public static Reader getDtd() {
    return new BufferedReader(new InputStreamReader(loader.getResourceAsStream("pluglist.dtd")));
  }
}
